// X36PAA_Batoh.cpp : Defines the entry point for the console application.
// Author: Marek Handl
// Date: December 30, 2008
// CTU, FEE, X36PAA
// Knapsack problem - Branch & Bound algorithm

/* Branch & Bound
 * at first use greedy algorithm to find one solution (recursion and always try to add item first - greedy)
 * then proceed with the rest of recursion
 * always compute theoretical maximum for each node (i.e. theoretical value if all resting items were added)
 * stop executing branches that have theoretical maximum lower than the best solution found so far (that's the lower bound)
 * naturaly stop executing branches that exceed the maximum allowed weight (that's the upper bound)
*/

#include "stdafx.h"
#include <vector>
#include <sstream>
#include <iostream>
#include <string>
#include <fstream>
#include "assert.h"
#include "process.h"
#include "windows.h"

using namespace std;


// ************************
// class representing one item
// ************************
class Item {
	int id;
	int price;
	int weight;
public:
	Item(int i, int p, int w) : id(i), price(p), weight(w) {}

	// getters
	int getPrice() {
		return price;
	}
	int getWeight() {
		return weight;
	}
};

// ************************
// class representing a knapsack
// ************************
class Knapsack {
	int maxWeight;
	int count;
	bool* content;
	int weight;
	int price;

public:
	Knapsack(int w, int c) : maxWeight(w), count(c), weight(0), price(0) {
		content = (count > 0 ? (bool*) malloc(count*sizeof(bool)) : NULL);
		for (int i=0; i<count; ++i) {
			content[i] = false;
		}
	}
	~Knapsack() {
		if (content != NULL) {
			free(content);
		}
	}
	Knapsack(const Knapsack& k) : maxWeight(k.maxWeight), count(k.count) {
		weight = k.weight;
		price = k.price;
		assert(count > 0);
		content = (count > 0 ? (bool*) malloc(count*sizeof(bool)) : NULL);
		for (int i=0; i<count; ++i) {
			content[i] = k.content[i];
		}
	}

	Knapsack& operator=(const Knapsack& k) {
		assert(count == k.count);
		maxWeight = k.maxWeight;
		weight = k.weight;
		price = k.price;
		for (int i=0; i<count; ++i) {
			assert(content != NULL);
			content[i] = k.content[i];
		}
		return *this;
	}
	// no need for operator= or copy constructor, everything is static

	// reset Knapsack (erase existing data, set new values)
	void reset(int w, int c) {
		maxWeight = w;
		weight = 0;
		price = 0;
		if (count != c) {	// in case count of items changes
			free(content);
			count = c;
			content = (count > 0 ? (bool*) malloc(count*sizeof(bool)) : NULL);
		}
		for (int i=0; i<count; ++i) {
			assert(content != NULL);
			content[i] = 0;
		}
	}
	
	// try to add an item into the sack
	// if it would exceed maxWeight, don't add and return false
	bool add(vector<Item> &items, int index) {
		assert(index < (int) items.size());
		if (weight + items.at(index).getWeight() > maxWeight) {	// maximum weight overflow
			return false;
		}
		price += items.at(index).getPrice();
		weight += items.at(index).getWeight();
		content[index] = true;
		return true;
	}

	// remove item from the sack
	void remove(vector<Item> &items, int index) {
		assert(index < (int) items.size());
		assert(content[index]);	// make sure the item is in the sack
		price -= items.at(index).getPrice();
		weight -= items.at(index).getWeight();
		content[index] = false;
	}

	// getter
	int getPrice() const {
		return price;
	}

	// print knapsack into a file
	void printResult(ofstream &fout, int instanceId=0) {
		try {
			fout << instanceId << " " << count << " " << price << " ";
			for (int i=0; i<count; ++i) {
				fout << " " << (int) content[i];
			}
			fout << endl;
		}
		catch (...) {
			cout << "Output file error" << endl;
			exit(1);
		}
	}
	

	// output for debugging
	void printOut() const {
		cout << "Weight: " << weight << " (" << maxWeight << ")" << endl;
		cout << "Price: " << price << endl;
		for (int i=0; i<count; ++i) {
			cout << (int) content[i] << " ";
		}
	}
};


// ************************
// read input data from a file
// ************************
bool readLineFromFile(ifstream &fin, int &id, int &count, int &maxWeight, vector<Item> &items) {
	string line;
	bool readMore = true;
	id = count = maxWeight = 0;

	try {		
		if (fin.rdstate() & fstream::failbit) {		// if file not found or unable to read from it
			throw "Input file not found";
		}
		
		if (!(fin >> id >> count >> maxWeight)) {
			readMore = false;
		}
		else {
			items.reserve(count);
			for (int i=0; i<count; ++i) {
				int weight, price;
				fin >> weight >> price;
				Item item(i, price, weight);
				items.push_back(item);
			}
		}
		
	}
	catch (...) {
		fin.close();
		cout << "Input file error" << endl;
		exit(1);
	}
	return readMore;
}

// ************************
// logic of the program
// ************************
void solve(Knapsack sack, vector<Item> &items, int index, Knapsack &best, int theoreticalMax, int &statesCount) {
	
	++statesCount;
	// first branch - add the item
	if (sack.add(items, index)) {	// try to add current item into the sack, succeeds only if sack weight would remain below maxWeight
		if (index < (int) items.size()-1) {
			solve(sack, items, index+1, best, theoreticalMax, statesCount);	// recursion call
		}
		if (sack.getPrice() > best.getPrice()) {		// find the best solution value
			best = sack;
		}
		sack.remove(items, index);	// remove the item to return to prior state
	}

	// second branch - don't add the item
	int newTheoreticalMax = theoreticalMax-items[index].getPrice();
	if (index < (int) items.size()-1 && newTheoreticalMax > best.getPrice()) {		// this is the bound from B&B
		solve(sack, items, index+1, best, newTheoreticalMax, statesCount);	// recursion call - do not add current item into the sack
	}
	if (sack.getPrice() > best.getPrice()) {		// find the best solution value
		best = sack;
	}
}


// ************************
// function used to measure time
// ************************
float GetMyCPUTime()
{
    HANDLE hProcess;    // specifies the process of interest
    FILETIME lpCreationTime;    // when the process was created
    FILETIME lpExitTime;    // when the process exited
    FILETIME lpKernelTime;    // time the process has spent in kernel mode
    FILETIME lpUserTime;
    LARGE_INTEGER t1, t2;


    hProcess = GetCurrentProcess();
    GetProcessTimes(hProcess, &lpCreationTime, &lpExitTime, &lpKernelTime, &lpUserTime);
    memcpy(&t1, &lpUserTime, sizeof(FILETIME));
    memcpy(&t2, &lpKernelTime, sizeof(FILETIME));
    return ((float)(t1.QuadPart+t2.QuadPart)/10000000);
}

// ************************
// ************************
int main(int argc, char* argv[])
{
	int instanceId;
	int count;
	int maxWeight;
	int statesTestedCount = 0;	// how many states get tested throughout the computation
	int statesCountOverall = 0;
	vector<Item> items;
	ifstream fin;
	ofstream fout;
	string outputFile("output.txt");
	string inputFile("input.txt");
	float time = 0, start, end;
	int instancesCount = 0;
	
	if (argc < 2) {		// make sure program is called with parameters
		cout << "Usage: X36PAA_Batoh.exe input_file [output_file]" << endl;
		return 0;
	}
	inputFile = string(argv[1]);
	outputFile = "v_"+inputFile;
	if (argc > 2) {		// second parameter is optional and defines output file
		outputFile = string(argv[2]);
	}	

	fin.open(inputFile.c_str(), fstream::in);	// read only access
	fout.open(outputFile.c_str(), fstream::out);	// write only access
	
	Knapsack sack(1, 1);
	Knapsack best(1, 1);

	while (readLineFromFile(fin, instanceId, count, maxWeight, items) ) {
		sack.reset(maxWeight, count);
		best.reset(maxWeight, count);
		
		// get theoretical max (sum costs of all items)
		int costSum = 0;
		for (int i=0; i<count; ++i) {
			costSum += items[i].getPrice();
		}

		start = GetMyCPUTime();
		solve(sack, items, 0, best,  costSum, statesTestedCount);	// compute solution
		end = GetMyCPUTime();

		time += end-start;
		++instancesCount;
		best.printResult(fout, instanceId);		// save result to a file
		items.clear();
		statesCountOverall += statesTestedCount;
		statesTestedCount = 0;
	}
	
	fin.close();
	fout.close();

	cout << "Instances: " << instancesCount << endl;
/*	cout << "Time: " << time << " sec" << endl;
	cout << "Time per instance: " << (float)time / (float)instancesCount << " sec" << endl;*/
	cout << "States tested per instance: " << statesCountOverall / instancesCount << endl;
	return 0;
}

