// Author: Marek Handl
// Date: December 30, 2008
// CTU, FEE, X36PAA
// Knapsack problem - heuristic based on price/weight ratio with most valued item test

// Algorithm
// find solution using heuristic based on price/weight ratio
// find solution that contains only the most valued item
// take the better solution

#include <vector>
#include <sstream>
#include <iostream>
#include <string>
#include <fstream>
#include "assert.h"
#include "process.h"
#include "windows.h"
#include <algorithm>
#include <math.h>

using namespace std;


// ************************
// class representing one item
// ************************
class Item {
	int id;
	int price;
	int weight;
public:
	// constructor
	Item(int i, int p, int w) : id(i), price(p), weight(w) {}

	// operator less than (used for sorting)
	bool operator<(const Item& i) const {
		return ((float)price/(float)weight < (float)(i.price)/(float)(i.weight));
	}

	// getters
	int getPrice() {
		return price;
	}
	int getWeight() {
		return weight;
	}
};

// ************************
// class representing a knapsack
// ************************
class Knapsack {
	int maxWeight;
	int count;
	bool* content;
	int weight;
	int price;

public:
	Knapsack(int w, int c) : maxWeight(w), count(c), weight(0), price(0) {
		content = (count > 0 ? (bool*) malloc(count*sizeof(bool)) : NULL);
		for (int i=0; i<count; ++i) {
			content[i] = false;
		}
	}
	
	~Knapsack() {
		if (content != NULL) {
			free(content);
		}
	}
	
	Knapsack(const Knapsack& k) : maxWeight(k.maxWeight), count(k.count) {
		weight = k.weight;
		price = k.price;
		assert(count > 0);
		content = (count > 0 ? (bool*) malloc(count*sizeof(bool)) : NULL);
		for (int i=0; i<count; ++i) {
			content[i] = k.content[i];
		}
	}

	Knapsack& operator=(const Knapsack& k) {
		assert(count == k.count);
		maxWeight = k.maxWeight;
		weight = k.weight;
		price = k.price;
		for (int i=0; i<count; ++i) {
			assert(content != NULL);
			content[i] = k.content[i];
		}
		return *this;
	}

	// reset Knapsack (erase existing data, set new values)
	void reset(int w, int c) {
		maxWeight = w;
		weight = 0;
		price = 0;
		if (count != c) {	// in case count of items changes
			free(content);
			count = c;
			content = (count > 0 ? (bool*) malloc(count*sizeof(bool)) : NULL);
		}
		for (int i=0; i<count; ++i) {
			assert(content != NULL);
			content[i] = 0;
		}
	}
	
	// try to add an item into the sack
	// if it would exceed maxWeight, don't add and return false
	bool add(vector<Item> &items, int index) {
		assert(index < (int) items.size());
		if (weight + items.at(index).getWeight() > maxWeight || items.at(index).getPrice() == 0) {	// maximum weight overflow
			return false;
		}
		price += items.at(index).getPrice();
		weight += items.at(index).getWeight();
		content[index] = true;
		return true;
	}

	// getter
	int getPrice() const {
		return price;
	}

	// print knapsack into a file
	void printResult(ofstream &fout, int instanceId=0) {
		try {
			fout << instanceId << " " << count << " " << price << " ";
			for (int i=0; i<count; ++i) {
				fout << " " << (int) content[i];
			}
			fout << endl;
		}
		catch (...) {
			cout << "Output file error" << endl;
			exit(1);
		}
	}
	

	// output for debugging
	void printOut() const {
		cout << "Weight: " << weight << " (" << maxWeight << ")" << endl;
		cout << "Price: " << price << endl;
		for (int i=0; i<count; ++i) {
			cout << (int) content[i] << " ";
		}
	}
};


// ************************
// read input data from a file
// ************************
bool readLineFromFile(ifstream &fin, int &id, int &count, int &maxWeight, vector<Item> &items) {
	string line;
	bool readMore = true;
	id = count = maxWeight = 0;

	try {		
		if (fin.rdstate() & fstream::failbit) {		// if file not found or unable to read from it
			throw "Input file not found";
		}
		
		if (!(fin >> id >> count >> maxWeight)) {
			readMore = false;
		}
		else {
			items.reserve(count);
			for (int i=0; i<count; ++i) {
				int weight, price;
				fin >> weight >> price;
				Item item(i, price, weight);
				items.push_back(item);
			}
		}
		
	}
	catch (...) {
		fin.close();
		cout << "Input file error" << endl;
		exit(1);
	}
	return readMore;
}


// ************************
// heuristic solution
// sorts the items based on price/weight ratio in descending order
// ************************
void solveHeuristic(Knapsack &sack, vector<Item> &items, int &statesTested) {
	sort(items.rbegin(), items.rend());
	for (int i=0; i<(int)items.size(); ++i) {
		++statesTested;
		if (!sack.add(items, i)) {
			break;
		}
	}
}

// get solution that contains only the most valued item that fits in the bag
void solveMostValued(Knapsack &sack, vector<Item> &items, const int maxWeight) {
	int temp;
	// solve it using heuristic without the test
	solveHeuristic(sack, items, temp);
	
	int indexOfBest = 0;
	// find the most valued item
	for (int i=1; i<(int)items.size(); ++i) {
		if (items[i].getWeight() <= maxWeight && items[i].getPrice() > items[indexOfBest].getPrice()) {
			indexOfBest = i;
		}
	}
	if (items[indexOfBest].getPrice() > sack.getPrice()) {
		sack.reset(maxWeight, (int)items.size());
		sack.add(items, indexOfBest);
	}
}

// ************************
// function used to measure time
// ************************
float GetMyCPUTime()
{
    HANDLE hProcess;    // specifies the process of interest
    FILETIME lpCreationTime;    // when the process was created
    FILETIME lpExitTime;    // when the process exited
    FILETIME lpKernelTime;    // time the process has spent in kernel mode
    FILETIME lpUserTime;
    LARGE_INTEGER t1, t2;


    hProcess = GetCurrentProcess();
    GetProcessTimes(hProcess, &lpCreationTime, &lpExitTime, &lpKernelTime, &lpUserTime);
    memcpy(&t1, &lpUserTime, sizeof(FILETIME));
    memcpy(&t2, &lpKernelTime, sizeof(FILETIME));
    return ((float)(t1.QuadPart+t2.QuadPart)/10000000);
}

// ************************
// ************************
int main(int argc, char* argv[])
{
	int instanceId;
	int count;
	int maxWeight;
	vector<Item> items;
	ifstream fin;
	ofstream fout, foutMostValued;
	string outputFile("output.txt");
	string outputFileMostValued("output1.txt");
	string inputFile("input.txt");
	float timeHeuristic = 0, timeMostValued = 0, start, end;
	int instancesCount = 0;
	int maxInstancesCount = 0;	// zero means unlimited
	float errorHeuristic = 0, maxErrorHeuristic = 0;
	int statesTested = 0, statesTestedOverall = 0;

	if (argc < 2) {		// make sure program called with parameters
		cout << "Usage: knapsack.exe input_file [output_file [output_file_heuristic [maxInstancesCount]]]" << endl;
		return 0;
	}
	inputFile = string(argv[1]);
	if (argc > 2) {		// second parameter is optional and defines output file
		outputFile = string(argv[2]);
	}
	outputFileMostValued = "v_"+inputFile;
	if (argc > 3) {		// second parameter is optional and defines output file
		outputFileMostValued = string(argv[3]);
	}
	if (argc > 4) {
		maxInstancesCount = atoi(argv[4]);
	}

	fin.open(inputFile.c_str(), fstream::in);	// read only access
	fout.open(outputFile.c_str(), fstream::out);	// write only access
	foutMostValued.open(outputFileMostValued.c_str(), fstream::out);	// write only access
	
	Knapsack sack(1, 1);
	Knapsack best(1, 1);


	while (readLineFromFile(fin, instanceId, count, maxWeight, items) ) {
		
		sack.reset(maxWeight, count);
		best.reset(maxWeight, count);

		start = GetMyCPUTime();
		solveHeuristic(best, items, statesTested);	// compute using simple heuristic
		end = GetMyCPUTime();

		timeHeuristic += end-start;
		best.printResult(fout, instanceId);		// save result to a file

		// heuristics with test
		sack.reset(maxWeight, count);
		start = GetMyCPUTime();
		solveMostValued(sack, items, maxWeight);		// compute using heuristic with test
		end = GetMyCPUTime();
		
		timeMostValued += end-start;
		sack.printResult(foutMostValued, instanceId);

		// comparison
		//assert(sack.getPrice() <= best.getPrice());
		assert(sack.getPrice() != 0);
		float currentError = (best.getPrice() - sack.getPrice()) / (float)(best.getPrice());
		errorHeuristic += currentError;
		maxErrorHeuristic = (currentError > maxErrorHeuristic ? currentError : maxErrorHeuristic);

		items.clear();
		++instancesCount;
		if (instancesCount == maxInstancesCount) {	// stop execution if desired number of instances already computed
			break;
		}

		// count of states computed
		// n*log(n) for sorting
		// statesTested for how many items were actually added
		// 1 for comparison with most valued item
		statesTestedOverall += count * (log((float)count)/log(2.0)) + statesTested + 1;
		statesTested = 0;
	}
	
	fin.close();
	fout.close();

	/*cout << "Instances: " << instancesCount << endl;
	cout << "*** Heuristic price/weight ***" << endl;
	cout << "Time: " << timeHeuristic << " sec" << endl;
	cout << "Time per instance: " << timeHeuristic / (float)instancesCount << " sec" << endl;
	cout << "*** Heuristic price/weight with test of most valued item ***" << endl;
	cout << "Time (heuristic): " << timeMostValued << " sec" << endl;
	cout << "Time per instance (heuristic): " << timeMostValued / (float)instancesCount << " sec" << endl;
	cout << "*** Comparison ***" << endl;
	cout << "Average errorHeuristic: " << errorHeuristic / (float)instancesCount * 100 << "%" << endl;
	cout << "Maximum errorHeuristic: " << maxErrorHeuristic*100 << "%" << endl;*/
	cout << "Tested states count: " << statesTestedOverall / instancesCount << endl;
	return 0;
}

